"""公共处理类"""


class BaseHandle:
    handle_key = None

    def __init__(
            self,
            func=None,
            func_params=None,
            limit_config=None,
            default_return=None
    ):
        """
        :param func: 被限流的方法
        :param func_params: 被限流的方法的参数
        :param limit_config: 限流配置信息
        :param default_return: 默认返回值【丢弃处理的时候需要返回值】
        """
        self.func = func
        self.func_params = func_params or {}
        self.limit_config = limit_config or {}
        self.default_return = default_return

    def _get_func_args(self):
        return self.func_params.get('args') or []

    def _get_func_kwargs(self):
        return self.func_params.get('kwargs') or {}

    def _get_handle_params(self):
        return self.limit_config.get('handle_params') or {}

    def _get_key_name(self):
        return self.limit_config.get('key_name')

    async def execute(self):
        """所有处理方式都要重写该方法"""
        raise NotImplementedError('未定义执行方法')



import base64
import dill as pickle
import aioredis_cluster

from ratelimit.base.constants import RedisConfig

try:
    from aioredis import create_redis_pool, ConnectionsPool
except:
    raise Exception('未安装 aioredis 库，支持 1.X.X 版本')


def encode_serialize(obj):
    """
    将对象序列化成字符串
    :param obj: 要序列化的对象
    :return: obj_str
    """
    return base64.b64encode(pickle.dumps(obj)).decode()


def decode_serialize(obj_str):
    """
    将字符串反序列化成对象
    :param obj_str: 被序列化的字符串
    :return: obj
    """
    try:
        return pickle.loads(base64.b64decode(obj_str))
    except Exception as err:
        return None


async def get_redis_connection():
    """获取一个redis连接"""
    if RedisConfig.conn:
        return RedisConfig.conn

    if RedisConfig.is_cluster:
        # 集群模式
        assert RedisConfig.startup_nodes, '集群节点未配置'
        RedisConfig.conn = await aioredis_cluster.create_redis_cluster(
            startup_nodes=RedisConfig.startup_nodes,
            password=RedisConfig.password,
            pool_minsize=RedisConfig.minsize,
            pool_maxsize=RedisConfig.maxsize,
            connect_timeout=RedisConfig.timeout,
            **(RedisConfig.kwargs or {})
        )
    else:
        # 单机模式
        RedisConfig.conn = await create_redis_pool(
            address=[RedisConfig.host, RedisConfig.port],
            db=RedisConfig.db,
            password=RedisConfig.password,
            minsize=RedisConfig.minsize,
            maxsize=RedisConfig.maxsize,
            timeout=RedisConfig.timeout,
            **(RedisConfig.kwargs or {})
        )

    return RedisConfig.conn



from ratelimit.handles import HandleFactory
from ratelimit.limits.redis_limit import RedisRateLimit


class OverLimitHandle:

    def __init__(self, func, func_params, key_name, default_return=None):
        """
        当流量超限时的处理类
        :param func: 超限的方法
        :param func_params: 超限方法的参数
        :param key_name: 获取超限配置的key
        :param default_return: 默认返回值
        """
        self.func = func
        self.func_params = func_params
        self.key_name = key_name
        self.default_return = default_return

    async def _can_do_func(self):
        """检查流量是否限制"""
        result, self.limit_config = await RedisRateLimit().attempt_get_token(key_name=self.key_name)
        return result

    async def _do_if_limit(self):
        """如果被限流了，根据配置信息做相应的处理"""
        handle_cls = HandleFactory().get_handle_cls(self.limit_config.get('handle'))
        return await handle_cls(
            func=self.func,
            func_params=self.func_params,
            limit_config=self.limit_config,
            default_return=self.default_return
        ).execute()

    async def _run_func(self):
        """运行监听的方法"""
        func_args = self.func_params.get('args') or []
        func_kwargs = self.func_params.get('kwargs') or {}
        return await self.func(*func_args, **func_kwargs)

    async def execute(self, exec_default_handle=True):
        """
        :param exec_default_handle: 是否执行默认操作
        :return:
        """
        if await self._can_do_func():
            return await self._run_func()

        # 如果不执行默认配置的操作，直接返回默认结果
        if not exec_default_handle:
            return self.default_return

        return await self._do_if_limit()



"""基于本地队列，可能会出现数据丢失的问题"""
import asyncio
import time
from asyncio import Queue

from ratelimit.limits.redis_limit import RedisRateLimit
from ratelimit.queue_tools.base import BaseQueue


class LocalQueueTools(BaseQueue):
    """本地内存队列工具类"""

    _queue = Queue()
    queue_type = 'local'

    async def add_item(self, item):
        """
        加入队列
        :param item: 这里队列的格式如下：
            {
                "func": func
                "func_args": [],
                "func_kwargs": {},
                "limit_config": {},
                "delay": 3
            }
        """
        delay = int(item.get('delay') or 0)
        item['handle_ts'] = int(time.time()) + delay
        await self._queue.put(item)

    async def get_item(self):
        """获取队列信息"""
        if not self._queue.qsize():
            return None

        item = await self._queue.get() or {}
        if item['handle_ts'] <= time.time():
            return item

        # 未到时间处理，则放回队列中
        item['delay'] = item['handle_ts'] - time.time()
        await self.add_item(item)

    async def _consume(self):
        """消费队列【内存队列，这里可能会存在任务丢失的情况】"""
        while True:
            item = await self.get_item()
            if not item:
                await asyncio.sleep(1)
                continue

            func = item.get('func')
            func_args = item.get('func_args')
            func_kwargs = item.get('func_kwargs')
            limit_config = item.get('limit_config')
            key_name = limit_config.get('key_name')

            # 这里也要判断是否超限
            result, _ = await RedisRateLimit().attempt_get_token(key_name, limit_config)
            if not result:
                # 若再次被限制时，延时执行时间修改为limit_delay
                handle_params = limit_config.get('handle_params') or {}
                item['delay'] = handle_params.get('limit_delay') or 0
                await self.add_item(item)
                continue

            await func(*func_args, **func_kwargs)
            await asyncio.sleep(0)



"""一些常量"""


class LimitConfig:
    # 限流器所属服务
    service = 'default'

    # 总令牌桶数量
    total_quota = 100

    # 单位时间内生成的令牌数
    limit_quota = 3

    # 单位时间
    limit_second = 1

    # 一次请求获取的令牌数
    once_quota = 1

    # 被限流时的处理方式，默认丢弃【discard丢弃；queue排队；retry重试】
    default_handle = 'discard'

    # 处理方式对应的默认配置参数
    default_handle_params = {}

    # 是否使用redis的配置，这里如果是True，会优先去redis中获取配置
    use_redis = False

    """
    详细配置，这里可以分别给每个具体的方法配置，具体格式如下：
    {
        "key_name": {
            "total_quota": 100,          # 令牌桶大小
            "limit_second": 10,          # 生成令牌的单位时间
            "limit_quota": 3,            # 单位时间内生成的令牌数量
            "once_quota": 1,             # 每次请求消费的令牌数量
            "handle": "discard",         # discard丢弃；queue排队；retry重试
            "handle_params": {},         # 处理方式对应的参数，详细见handles的注释
        }
    }
    """
    rate_limit_config = {}


class RedisConfig:
    # 地址
    host = '127.0.0.1'

    # 端口
    port = 6379

    # 数据库
    db = 0

    # 密码
    password = None

    # 最小连接数
    minsize = 1

    # 最大连接数
    maxsize = 10

    # 超时时间【毫秒】
    timeout = 5000

    # 额外参数
    kwargs = None

    # 初始化连接池
    conn = None

    # 是否使用集群方式
    is_cluster = False

    """
    redis集群节点配置，is_cluster为true时需要配置，格式如下：
    [
        ["127.0.0.1", 6379]
    ]
    """
    startup_nodes = []



def singleton(cls_):
    """单例类装饰器"""

    class wrap_cls(cls_):
        __instance = None

        def __new__(cls, *args, **kwargs):
            if cls.__instance is None:
                cls.__instance = super().__new__(cls, *args, **kwargs)
                cls.__instance.__init = False
            return cls.__instance

        def __init__(self, *args, **kwargs):
            if self.__init:
                return
            super().__init__(*args, **kwargs)
            self.__init = True

    wrap_cls.__name__ = cls_.__name__
    wrap_cls.__doc__ = cls_.__doc__
    wrap_cls.__qualname__ = cls_.__qualname__
    return wrap_cls



from ratelimit.queue_tools import QueueFactory
from ratelimit.handles.base import BaseHandle


class QueueHandle(BaseHandle):
    """排队处理"""
    handle_key = 'queue'

    async def execute(self):
        """
        排队处理，配置信息 self.handle_params 格式如下：
            {
                "delay": 3,             # 延时执行时间，单位【秒】
                "limit_delay": 1,         # 执行再次被限制的情况下，延迟执行时间，单位【秒】
                "queue_type": "redis",  # 队列类型：redis、local
            }
        """
        func_args = self._get_func_args()
        func_kwargs = self._get_func_kwargs()
        handle_params = self._get_handle_params()

        queue_type = handle_params.get('queue_type') or 'local'
        delay = handle_params.get('delay') or 0

        item = {
            'func': self.func,
            'func_args': func_args,
            'func_kwargs': func_kwargs,
            'limit_config': self.limit_config,
            'delay': delay
        }

        queue_cls = QueueFactory().get_queue_cls(queue_type)
        await queue_cls().add_item(item)



from ratelimit.base.base_decorators import singleton
from ratelimit.base.errors import NotDefineHandleError, NotAllowHandleError
from ratelimit.handles.base import BaseHandle
from ratelimit.handles.discard_handle import DiscardHandle
from ratelimit.handles.queue_handle import QueueHandle
from ratelimit.handles.retry_handle import RetryHandle


@singleton
class HandleFactory:
    """处理工厂类"""
    handle_dict = {}

    def __init__(self):
        self._register(DiscardHandle)
        self._register(QueueHandle)
        self._register(RetryHandle)

    def _register(self, handle_cls):
        """把处理方法注册进来"""
        if not issubclass(handle_cls, BaseHandle):
            raise NotAllowHandleError(f'处理类【{handle_cls}】不允许注册')

        self.handle_dict[handle_cls.handle_key] = handle_cls

    def get_handle_cls(self, handle_key):
        """获取真正的处理类"""
        handle_cls = self.handle_dict.get(handle_key)
        if not handle_cls:
            raise NotDefineHandleError(f'未定义【{handle_key}】的处理类')

        return handle_cls



"""队列操作基础类"""
import asyncio

from ratelimit.base.base_decorators import singleton


@singleton
class BaseQueue:
    _has_thread = False

    def __init__(self):
        self._create_task()

    async def add_item(self, item):
        """
        加入队列
        :param item: 这里队列的格式如下：
            {
                "func": func
                "func_args": [],
                "func_kwargs": {},
                "limit_config": {},
                "delay": 3
            }
        """
        raise NotImplementedError('未定义【add_item】方法')

    async def get_item(self):
        """获取队列信息"""
        raise NotImplementedError('未定义【get_item】方法')

    async def _consume(self):
        """消费队列"""
        raise NotImplementedError('未定义【_consume】方法')

    def _create_task(self):
        """起一个协程消费"""
        if not self._has_thread:
            self._has_thread = True
            # 将消费函数函数注册到 event_loop 上
            _loop = asyncio.get_event_loop()
            asyncio.run_coroutine_threadsafe(self._consume(), _loop)




class LimitError(Exception):
    """限流器出错"""


class NotDefaultValueError(LimitError):
    """缺少默认返回值"""


class LackParamsError(LimitError):
    """缺少参数信息"""


class NotAllowHandleError(LimitError):
    """不是一个允许的处理类"""


class NotDefineHandleError(LimitError):
    """未定义处理类"""


class ReWriteStopIteration(Exception):
    """重写StopIteration报错"""



"""基于redis-cell的流量控制器"""
import asyncio
import logging
import traceback

import ujson as json

from ratelimit.base.base_decorators import singleton
from ratelimit.base.constants import LimitConfig
from ratelimit.base.tools import get_redis_connection

logger = logging.getLogger(__name__)


@singleton
class RedisRateLimit:

    async def get_limit_config(self, key_name):
        """
        获取key对应的配置【直接把限流配置信息存在redis中，直接获取】
        :param key_name: redis键名
        :return: 限流配置，如果是用了redis的配置【使用hash类型】，在redis中的存储格式如下
            "service": {
                "key_name": {
                    "total_quota": 100,          # 令牌桶大小
                    "limit_second": 10,          # 生成令牌的单位时间
                    "limit_quota": 3,            # 单位时间内生成的令牌数量
                    "once_quota": 1,             # 每次请求消费的令牌数量
                    "handle": "discard",         # discard丢弃；queue排队；retry重试
                    "handle_params": {},         # 处理方式对应的参数
                }
            }
        """
        limit_config = LimitConfig.rate_limit_config.get(key_name) or {}
        if LimitConfig.use_redis:
            conn = await get_redis_connection()
            limit_config = json.loads(await conn.hget(LimitConfig.service, key_name) or '{}')

        limit_config['service'] = LimitConfig.service
        limit_config['key_name'] = key_name
        limit_config['total_quota'] = limit_config.get('total_quota') or LimitConfig.total_quota
        limit_config['limit_quota'] = limit_config.get('limit_quota') or LimitConfig.limit_quota
        limit_config['limit_second'] = limit_config.get('limit_second') or LimitConfig.limit_second
        limit_config['once_quota'] = limit_config.get('once_quota') or LimitConfig.once_quota
        limit_config['handle'] = limit_config.get('handle') or LimitConfig.default_handle
        limit_config['handle_params'] = limit_config.get('handle_params') or LimitConfig.default_handle_params

        return limit_config

    async def attempt_get_token(self, key_name, limit_config=None):
        """
        尝试获取令牌，如果获取成功，返回true，代表可访问；
        :param key_name: redis键名
        :param limit_config: 配置信息，不传则从redis中获取
        :return: (bool, dict) -> (是否允许通过, 限流配置信息)
        """
        try:
            if not limit_config:
                limit_config = await self.get_limit_config(key_name)

            total_quota = limit_config.get('total_quota') or LimitConfig.total_quota
            limit_quota = limit_config.get('limit_quota') or LimitConfig.limit_quota
            limit_second = limit_config.get('limit_second') or LimitConfig.limit_second
            once_quota = limit_config.get('once_quota') or LimitConfig.once_quota

            conn = await get_redis_connection()
            result = await conn.execute(
                'CL.THROTTLE',
                f'{LimitConfig.service}:{key_name}',
                total_quota,
                limit_quota,
                limit_second,
                once_quota
            )
            logger.debug(f'* limit result【{LimitConfig.service} - {key_name}】', result)
            return result[0] == 0, limit_config
        except Exception as err:
            # 出现报错直接允许【高可用】
            logger.warning(f'尝试获取令牌出错：{traceback.format_exc()}')
            return True, {}


if __name__ == '__main__':
    redis_rate_limit = RedisRateLimit()
    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(redis_rate_limit.attempt_get_token('key_name'))
    print(loop.run_until_complete(task))



import asyncio

from ratelimit.base.errors import NotDefaultValueError
from ratelimit.handles.base import BaseHandle
from ratelimit.limits.redis_limit import RedisRateLimit


class RetryHandle(BaseHandle):
    """流量超限做重试处理"""
    handle_key = 'retry'

    async def execute(self):
        """
        重试处理，配置信息 self.handle_params 格式如下：
            {
                "retry_count"：3,    # 重试次数
                "retry_wait": 0,     # 每次重试间隔，默认不等待，单位【秒】
                "current_count": 0,  # 当前已重试几次
            }
        """
        func_args = self._get_func_args()
        func_kwargs = self._get_func_kwargs()
        handle_params = self._get_handle_params()
        key_name = self._get_key_name()

        current_count = 0
        retry_count = handle_params.get('retry_count') or 1
        retry_wait = handle_params.get('retry_wait') or 0
        while current_count < retry_count:
            await asyncio.sleep(retry_wait)
            # 这里也要判断是否超限
            result, _ = await RedisRateLimit().attempt_get_token(key_name, self.limit_config)
            if not result:
                current_count += 1
                continue
            return await self.func(*func_args, **func_kwargs)

        # 重试失败，则返回默认值
        if self.default_return is None:
            raise NotDefaultValueError('重试失败，缺少默认返回值')
        return self.default_return



from ratelimit.base.base_decorators import singleton
from ratelimit.base.errors import NotDefineHandleError, NotAllowHandleError
from ratelimit.queue_tools.base import BaseQueue
from ratelimit.queue_tools.local_queue import LocalQueueTools
from ratelimit.queue_tools.redis_queue import RedisQueueTools


@singleton
class QueueFactory:
    """队列工厂类"""
    queue_dict = {}

    def __init__(self):
        self._register(LocalQueueTools)
        self._register(RedisQueueTools)

    def _register(self, queue_cls):
        """把队列方法注册进来"""
        if not issubclass(queue_cls, BaseQueue):
            raise NotAllowHandleError(f'队列类【{queue_cls}】不允许注册')

        self.queue_dict[queue_cls.queue_type] = queue_cls

    def get_queue_cls(self, queue_type):
        """获取真正的队列类"""
        queue_cls = self.queue_dict.get(queue_type)
        if not queue_cls:
            raise NotDefineHandleError(f'未定义【{queue_type}】的队列类')

        return queue_cls



"""基于redis队列，使用pickle进行序列化，不支持跨语言"""
import asyncio
import time

from ratelimit.base.constants import LimitConfig
from ratelimit.base.tools import encode_serialize, get_redis_connection, decode_serialize
from ratelimit.limits.redis_limit import RedisRateLimit
from ratelimit.queue_tools.base import BaseQueue


class RedisQueueTools(BaseQueue):
    """redis队列工具类"""

    queue_key = f'{LimitConfig.service}:limit:queue'
    queue_type = 'redis'

    async def add_item(self, item):
        """
        加入队列，使用定时队列处理
        :param item: 这里队列的格式如下：
            {
                "func": func
                "func_args": [],
                "func_kwargs": {},
                "limit_config": {},
                "delay": 3
            }
        """
        delay = int(item.get('delay') or 0)
        score = int(time.time()) + delay
        conn = await get_redis_connection()
        item['score'] = score
        await conn.zadd(
            key=self.queue_key,
            score=score,
            member=encode_serialize(item)
        )

    async def get_item(self):
        """
        获取队列信息，这里会返回多条数据
        :return:
        """
        conn = await get_redis_connection()
        min_score = 0
        max_score = time.time()

        pl = conn.pipeline()
        pl.zrangebyscore(
            key=self.queue_key,
            min=min_score,
            max=max_score
        )
        pl.zremrangebyscore(self.queue_key, min=min_score, max=max_score)
        item_list, _ = await pl.execute()
        return item_list

    async def _consume(self):
        """消费队列"""
        while True:
            item_list = await self.get_item()
            if not item_list:
                await asyncio.sleep(1)
                continue

            for obj_str in item_list:
                item = decode_serialize(obj_str)
                if not item:
                    continue

                func = item.get('func')
                func_args = item.get('func_args')
                func_kwargs = item.get('func_kwargs')
                limit_config = item.get('limit_config')
                key_name = limit_config.get('key_name')

                # 这里也要判断是否超限
                result, _ = await RedisRateLimit().attempt_get_token(key_name, limit_config)
                if not result:
                    # 若再次被限制时，延时执行时间修改为limit_delay
                    handle_params = limit_config.get('handle_params') or {}
                    item['delay'] = handle_params.get('limit_delay') or 0
                    await self.add_item(item)
                    continue

                await func(*func_args, **func_kwargs)
                await asyncio.sleep(0)



import setuptools
from os import path as os_path

this_directory = os_path.abspath(os_path.dirname(__file__))


def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()


def read_requirements(filename):
    return [
        line.strip() 
        for line in read_file(filename).splitlines() 
        if not line.startswith('#')
    ]


setuptools.setup(
    name="rate_limit_async",
    version="1.0.1",
    author="Yang Peng",
    author_email="yp_rocfly@foxmail.com",
    description="distributed rate limiter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yprocfly/rate-limit-async",
    keywords=['rate', 'rate-limit', 'limit', 'redis-cell'],
    packages=setuptools.find_packages(),
    install_requires=read_requirements('requirements.txt'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)



from ratelimit.base.errors import NotDefaultValueError
from ratelimit.handles.base import BaseHandle


class DiscardHandle(BaseHandle):
    """丢弃处理"""
    handle_key = 'discard'

    async def execute(self):
        """直接返回默认值"""
        if self.default_return is None:
            raise NotDefaultValueError('缺少默认返回值')
        return self.default_return



"""一些装饰器"""
import asyncio

from ratelimit.utils.over_limit_handle import OverLimitHandle


def async_rate_limit(key_name, default_return=None):
    """
    异步方法限流装饰器
    :param key_name: 键名，必传，理论上同一个系统中要唯一
    :param default_return: 默认返回值，不传则会抛出异常
    """
    class RateLimitDecorator:
        def __init__(self, func):
            self.func = func

        async def __call__(self, *args, **kwargs):
            return await OverLimitHandle(
                func=self.func,
                func_params={
                    'args': args,
                    'kwargs': kwargs,
                },
                key_name=key_name,
                default_return=default_return
            ).execute()

    return RateLimitDecorator


if __name__ == '__main__':
    async def test_limit():
        @async_rate_limit('test')
        async def test(number):
            from datetime import datetime
            print('*********', datetime.now(), number)

        import random
        from ratelimit.base.constants import LimitConfig
        LimitConfig.total_quota = 0
        LimitConfig.limit_quota = 1
        LimitConfig.default_handle = 'queue'
        LimitConfig.default_handle_params = {
            'delay': 3,
            'limit_delay': 1,
            'queue_type': 'local'
        }

        await test(random.random())
        await test(random.random())

        await asyncio.sleep(5)

        await test(random.random())
        print('----------------3')
        await test(random.random())
        await test(random.random())
        print('----------------4')
        await asyncio.sleep(10)

    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(test_limit())
    loop.run_until_complete(task)
